"""clean operation."""
from __future__ import annotations

import itertools
import shutil
from typing import TYPE_CHECKING

from result import Ok, Result

if TYPE_CHECKING:
    from pyrgo.core.models.config import Config


def execute(app_config: Config) -> Result[None, Exception]:
    """Execute clean operation."""
    for to_be_clean in itertools.chain(
        app_config.caches,
        app_config.artifacts,
    ):
        if to_be_clean.exists() and to_be_clean.is_dir():
            shutil.rmtree(to_be_clean)

    for relevant_path in app_config.relevant_paths:
        for pycaches in app_config.cwd.joinpath(relevant_path).rglob(
            pattern="__pycache__",
        ):
            shutil.rmtree(pycaches)

    for env_not_as_opt_dep in set(app_config.available_envs).difference(
        set(app_config.dependency_groups),
    ):
        app_config.requirements_dir.joinpath(
            f"{env_not_as_opt_dep}.{app_config.lock_file_format}",
        ).unlink()

    return Ok()



"""add command."""
from __future__ import annotations

import click


@click.command
def add() -> None:
    """
    Add dependencies to a pyproject.toml manifest file and install.

    \b
    Install packages from:
    - PyPI (and other indexes) using requirement specifiers.
    - VCS project urls.
    - Local project directories.
    - Local or remote source archives.
    """  # noqa: D301
    raise NotImplementedError



"""Test command models."""
from __future__ import annotations

import itertools
import sys

import pytest

from pyrgo.core.models.command import PythonExecCommand


@pytest.mark.integration()
def test_execute_command() -> None:
    """Test execute command."""
    PythonExecCommand(program="ruff").add_args(
        args=["--version"],
    ).run()


@pytest.mark.unit()
@pytest.mark.parametrize(
    argnames=["program", "commands"],
    argvalues=[
        (
            "program",
            [
                ["command-1", "command-2"],
            ],
        ),
        (
            "program",
            [
                ["command-1"],
                ["command-2"],
            ],
        ),
    ],
)
def test_args_building(program: str, commands: list[list[str]]) -> None:
    """Test args building works as expected."""
    command = PythonExecCommand(program=program)
    for set_of_commands in commands:
        command.add_args(args=set_of_commands)

    assert command.args == [
        sys.executable,
        "-m",
        program,
        *itertools.chain.from_iterable(commands),
    ]



"""Application contants."""
from __future__ import annotations

from pathlib import Path

from pyrgo.core.models.config import build_config

app_config = build_config(
    cwd=Path().cwd(),
    cache_directories=[
        ".mypy_cache",
        ".pytest_cache",
        ".ruff_cache",
    ],
    artifacts_directories=[
        "dist",
        "site",
    ],
    venv_dir=".venv",
    lock_file_format="txt",
    core_deps_alias="core",
)



"""test command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.core import ops
from pyrgo.core.constants import app_config


@click.command()
@click.option(
    "-m",
    "--marked",
    "marker",
    type=click.Choice(
        choices=app_config.pytest_makers,
    ),
    default=None,
    help="Run tests based on marks. By default all tests are executed.",
)
def test(
    *,
    marker: str | None,
) -> None:
    """Execute tests using `pytest`."""
    executed = ops.test.execute(
        marker=marker,
        app_config=app_config,
    )
    if not isinstance(executed, Ok):
        sys.exit(1)
    sys.exit(0)



"""docs new operation."""
from __future__ import annotations

from typing import TYPE_CHECKING

from pyrgo.core.models.command import (
    PythonExecCommand,
)
from pyrgo.core.utilities.command import inform_and_run_program

if TYPE_CHECKING:
    import subprocess

    from result import Result

    from pyrgo.core.models.config import Config


def execute(app_config: Config) -> Result[None, list[subprocess.CalledProcessError]]:
    """Execute docs new operation."""
    new_command = PythonExecCommand(
        program="mkdocs",
    ).add_args(
        args=[
            "new",
        ],
    )
    new_command.add_args(args=[app_config.cwd.as_posix()])
    return inform_and_run_program(
        commands=[new_command],
    )



"""Pyrgo cli 'python -m pyrgo.cli.pyrgo_core' entrypoint."""
from __future__ import annotations

import sys

if __name__ == "__main__":
    from pyrgo.cli.pyrgo_new.cli import new

    sys.exit(new())



"""Python command."""
from __future__ import annotations

import subprocess
import sys
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from result import Err, Ok, Result

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(
    eq=False,
    frozen=False,
    repr=True,
)
class PythonExecCommand:
    """Command builder."""

    program: str
    args: list[str] = field(init=False)

    def __post_init__(self) -> None:  # noqa: D105
        self.args = [sys.executable, "-m", self.program]

    def add_args(self, args: list[str]) -> Self:
        """Add args to execution."""
        self.args.extend(args)
        return self

    def run(self) -> Result[None, subprocess.CalledProcessError]:
        """Run command."""
        try:
            subprocess.run(args=self.args, check=True)
        except subprocess.CalledProcessError as e:
            return Err(e)

        return Ok()



"""docs command."""
from __future__ import annotations

import click

from pyrgo.cli.pyrgo_core.commands.docs.build import build
from pyrgo.cli.pyrgo_core.commands.docs.new import new
from pyrgo.cli.pyrgo_core.commands.docs.serve import serve
from pyrgo.cli.utils import add_commands


@click.group()
def docs() -> None:
    """Document you project with `mkdocs-material`."""


add_commands(
    cli=docs,
    commands=[
        serve,
        build,
        new,
    ],
)



"""Test command utilities."""
from __future__ import annotations

import pytest

from pyrgo.core.models.command import PythonExecCommand
from pyrgo.core.utilities.command import inform_and_run_program


@pytest.mark.integration()
def test_inform_and_run_program() -> None:
    """Test inform and run program."""
    inform_and_run_program(
        commands=[
            PythonExecCommand(program="ruff").add_args(
                args=["--version"],
            ),
        ],
    ).unwrap()


@pytest.mark.integration()
def test_inform_and_run_program_fails() -> None:
    """Test inform and run program."""
    execution_err = inform_and_run_program(
        commands=[
            PythonExecCommand(program="ruff").add_args(
                args=["-abc"],
            ),
        ],
    ).err()
    assert execution_err
    assert len(execution_err) == 1



"""Test errors."""
from __future__ import annotations

import pathlib

import pytest

from pyrgo.core.errors import PyProjectTOMLNotFoundError


@pytest.mark.unit()
class TestPyProjectNotFoundError:
    """Test `PyProjectNotFoundError`."""

    def test_error_message(self) -> None:
        """Test error message."""
        path_to_toml = pathlib.Path.cwd().joinpath("pyproject_toml")
        assert (
            str(PyProjectTOMLNotFoundError(pyproject_toml=path_to_toml))
            == f"No `pyproject.toml` found at current directory {path_to_toml!s}"
        )



"""lock command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.core import ops
from pyrgo.core.constants import app_config


@click.command()
@click.option(
    "-g",
    "--group",
    "groups",
    type=click.Choice(choices=app_config.dependency_groups),
    multiple=True,
    required=False,
    help="Name of an extras_require group to install; may be used more than once",
)
@click.option(
    "--generate-hashes",
    "generate_hashes",
    is_flag=True,
    default=False,
    show_default=True,
    help="Generate pip 8 style hashes in the resulting requirements file.",
)
def lock(*, groups: tuple[str, ...], generate_hashes: bool) -> None:
    """Lock dependencies using `piptools`."""
    executed = ops.lock.execute(
        groups=groups,
        app_config=app_config,
        generate_hashes=generate_hashes,
    )
    if not isinstance(executed, Ok):
        sys.exit(1)
    sys.exit(0)



"""lock operation."""
from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

from pyrgo.core.models.command import (
    PythonExecCommand,
)
from pyrgo.core.utilities.command import inform_and_run_program
from pyrgo.core.utilities.text import path_to_lock_file

if TYPE_CHECKING:
    import subprocess

    from result import Result

    from pyrgo.core.models.config import Config


def execute(
    *,
    groups: Iterable[str],
    app_config: Config,
    generate_hashes: bool,
) -> Result[None, list[subprocess.CalledProcessError]]:
    """Execute lock operation."""
    app_config.requirements_dir.relative_to(app_config.cwd)
    commands: list[PythonExecCommand] = []

    if not groups:
        groups = app_config.dependency_groups

    for group in groups:
        piptools_command = PythonExecCommand(
            program="piptools",
        ).add_args(
            args=[
                "compile",
            ],
        )

        if group != app_config.core_deps_alias:
            piptools_command.add_args(
                args=[
                    "--extra",
                    group,
                ],
            )

        lock_file_path = path_to_lock_file(
            cwd=app_config.cwd,
            requirements_path=app_config.requirements_dir,
            group=group,
            lock_file_format=app_config.lock_file_format,
        )
        if generate_hashes:
            piptools_command.add_args(
                args=[
                    "--generate-hashes",
                ],
            )

        piptools_command.add_args(
            args=[
                "--resolver=backtracking",
                "-o",
                lock_file_path,
                "pyproject.toml",
            ],
        )
        commands.append(piptools_command)

    return inform_and_run_program(
        commands=commands,
    )



"""audit operation."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import subprocess

    from result import Result

    from pyrgo.core.models.config import Config

from pyrgo.core.models.command import PythonExecCommand
from pyrgo.core.utilities.command import inform_and_run_program
from pyrgo.core.utilities.text import path_to_lock_file


def execute(
    *,
    app_config: Config,
    environment: str,
    fix: bool,
) -> Result[None, list[subprocess.CalledProcessError]]:
    """Execute audit operation."""
    pipaudit_command = PythonExecCommand(
        program="pip_audit",
    )

    lock_file_path = path_to_lock_file(
        cwd=app_config.cwd,
        requirements_path=app_config.requirements_dir,
        group=environment,
        lock_file_format=app_config.lock_file_format,
    )

    pipaudit_command.add_args(
        args=[
            "-r",
            lock_file_path,
            "-l",
            "--desc",
        ],
    )
    if fix:
        pipaudit_command.add_args(
            args=[
                "--fix",
            ],
        )

    return inform_and_run_program(
        commands=[
            pipaudit_command,
        ],
    )



"""fmt operation."""
from __future__ import annotations

from typing import TYPE_CHECKING

from pyrgo.core.models.command import (
    PythonExecCommand,
)
from pyrgo.core.utilities.command import inform_and_run_program

if TYPE_CHECKING:
    import subprocess

    from result import Result

    from pyrgo.core.models.config import Config


def execute(app_config: Config) -> Result[None, list[subprocess.CalledProcessError]]:
    """Execute fmt operation."""
    ruff_command = PythonExecCommand(program="ruff").add_args(args=["--fix-only"])
    black_command = PythonExecCommand(program="black")

    for command in [ruff_command, black_command]:
        command.add_args(
            args=app_config.relevant_paths,
        )

    return inform_and_run_program(
        commands=[
            ruff_command,
            black_command,
        ],
    )



"""audit command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.core import ops
from pyrgo.core.constants import app_config


@click.command
@click.option(
    "-e",
    "--env",
    "environment",
    type=click.Choice(choices=app_config.available_envs),
    required=True,
    help="Audit to one of available enviroments.",
)
@click.option(
    "--fix",
    "fix",
    is_flag=True,
    show_default=True,
    default=False,
    help="Automatically upgrade dependencies with known vulnerabilities",
)
def audit(*, environment: str, fix: bool) -> None:
    """
    Scan project requirements for packages with known vulnerabilities.

    It uses the Python Packaging Advisory Database (https://github.com/pypa/advisory-database)
    """
    executed = ops.audit.execute(
        environment=environment,
        app_config=app_config,
        fix=fix,
    )
    if not isinstance(executed, Ok):
        sys.exit(1)

    sys.exit(0)



"""test operation."""
from __future__ import annotations

from typing import TYPE_CHECKING

from pyrgo.core.models.command import (
    PythonExecCommand,
)
from pyrgo.core.utilities.command import inform_and_run_program

if TYPE_CHECKING:
    import subprocess

    from result import Result

    from pyrgo.core.models.config import Config


def execute(
    *,
    marker: str | None,
    app_config: Config,
) -> Result[None, list[subprocess.CalledProcessError]]:
    """Execute test operation."""
    pytest_command = PythonExecCommand(
        program="pytest",
    )
    if marker:
        pytest_command.add_args(
            args=[
                "-m",
                marker,
            ],
        )

    pytest_command.add_args(args=app_config.pytest_paths)
    return inform_and_run_program(
        commands=[
            pytest_command,
        ],
    )



"""docs serve operation."""
from __future__ import annotations

from typing import TYPE_CHECKING

from pyrgo.core.models.command import (
    PythonExecCommand,
)
from pyrgo.core.utilities.command import inform_and_run_program

if TYPE_CHECKING:
    import subprocess

    from result import Result


def execute(
    *,
    dev_address: str,
    theme: str,
    strict: bool,
) -> Result[None, list[subprocess.CalledProcessError]]:
    """Execute docs serve operation."""
    serve_command = PythonExecCommand(
        program="mkdocs",
    ).add_args(
        args=[
            "serve",
        ],
    )
    if strict:
        serve_command.add_args(args=["--strict"])

    serve_command.add_args(
        args=[
            "--theme",
            theme,
            "--dev-addr",
            dev_address,
        ],
    )

    return inform_and_run_program(
        commands=[serve_command],
    )



"""check command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.core import ops
from pyrgo.core.constants import app_config


@click.command()
@click.option(
    "--add-noqa",
    "add_noqa",
    is_flag=True,
    default=False,
    type=bool,
    help="Enable automatic additions of `noqa` directives to failing lines",
)
@click.option(
    "--ignore-noqa",
    "ignore_noqa",
    is_flag=True,
    default=False,
    type=bool,
    help="Ignore any `# noqa` comments",
)
def check(*, add_noqa: bool, ignore_noqa: bool) -> None:
    """Analyze the current package with `ruff` and `mypy`."""
    executed = ops.check.execute(
        add_noqa=add_noqa,
        ignore_noqa=ignore_noqa,
        app_config=app_config,
    )
    if not isinstance(executed, Ok):
        sys.exit(1)
    sys.exit(0)



"""Application errors."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


class PyProjectTOMLNotFoundError(Exception):
    """Raised when no `pyproject.toml` found."""

    def __init__(self, pyproject_toml: Path) -> None:
        """No `pyproject.toml` found in current directory."""
        super().__init__(
            f"No `pyproject.toml` found at current directory {pyproject_toml!s}",
        )



"""Root init CLI."""
from __future__ import annotations

import sys
import tempfile
from pathlib import Path

import click
from result import Ok

from pyrgo.core.utilities.io import prepare_starter_project


@click.command(
    context_settings={
        "help_option_names": [
            "-h",
            "--help",
        ],
    },
)
@click.argument("path", type=click.Path())
@click.option(
    "-n",
    "--name",
    "name",
    type=click.STRING,
    required=True,
    help="New project name.",
)
def new(path: str, name: str) -> None:
    """Initialize a new Pyrgo project."""
    path_as_path = Path(path)
    with tempfile.TemporaryDirectory() as tmp:
        template_path = prepare_starter_project(
            project_name=name,
            tmp_dir=Path(tmp),
        )
        if not isinstance(template_path, Ok):
            click.echo(template_path.err())
            sys.exit(1)

        if not path_as_path.exists():
            path_as_path.mkdir()

        template_path.unwrap().rename(path_as_path.absolute())



from __future__ import annotations

from pyrgo.core.ops import (
    audit,
    build,
    check,
    clean,
    docs,
    fmt,
    lock,
    sync,
    test,
    venv,
)

__all__ = [
    "test",
    "venv",
    "docs",
    "build",
    "fmt",
    "check",
    "lock",
    "sync",
    "clean",
    "audit",
]



"""Project config."""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import tomli

from pyrgo.core.errors import PyProjectTOMLNotFoundError

if TYPE_CHECKING:
    import sys
    from pathlib import Path

    if sys.version_info >= (3, 9):
        from typing import Literal
    else:
        from typing_extensions import Literal


@dataclass(
    frozen=True,
    repr=True,
    eq=False,
)
class Config:
    """App configuration setting."""

    cwd: Path
    requirements_dir: Path
    pyproject_file: Path
    caches: list[Path]
    artifacts: list[Path]
    venv_dir: Path
    venv_activation_msg: str
    core_deps_alias: str
    lock_file_format: Literal["txt"]
    pyproject: dict[str, Any]

    @property
    def available_envs(self) -> list[str]:
        """List available envs found at `requirements_dir`."""
        requirements_file_pattern = f"*.{self.lock_file_format}"
        return [
            x.name.rstrip(requirements_file_pattern)
            for x in self.requirements_dir.glob(
                pattern=requirements_file_pattern,
            )
            if x.is_file()
        ]

    @property
    def pytest_makers(self) -> list[str]:
        """List all custom-defined `pytest` markers."""
        makers: list[str] = self.pyproject["tool"]["pytest"]["ini_options"]["markers"]
        return [x.split(sep=":")[0] for x in makers]

    @property
    def optional_dependency_groups(self) -> dict[str, list[str]]:
        """List all optional-dependency groups."""
        return self.pyproject["project"]["optional-dependencies"]

    @property
    def dependency_groups(self) -> list[str]:
        """List all defined dependency groups."""
        dependency_groups: list[str] = [self.core_deps_alias]
        dependency_groups.extend(self.optional_dependency_groups)
        return dependency_groups

    @property
    def pytest_paths(self) -> list[str]:
        """List all `pytest` testpaths."""
        return self.pyproject["tool"]["pytest"]["ini_options"]["testpaths"]

    @property
    def project_path(self) -> str:
        """Get project name."""
        return self.pyproject["project"]["name"]

    @property
    def relevant_paths(self) -> list[str]:
        """Project relevant paths."""
        relevant_paths: list[str] = []
        relevant_paths.extend(self.pytest_paths)
        relevant_paths.append(self.project_path)
        return relevant_paths


def build_config(  # noqa: PLR0913
    cwd: Path,
    cache_directories: list[str],
    artifacts_directories: list[str],
    venv_dir: str,
    lock_file_format: Literal["txt"],
    core_deps_alias: str,
) -> Config:
    """Build config."""
    path_to_pyproject_toml = cwd.joinpath("pyproject.toml")
    if not path_to_pyproject_toml.exists() and path_to_pyproject_toml.is_file():
        raise PyProjectTOMLNotFoundError(pyproject_toml=path_to_pyproject_toml)

    requirements_path = cwd.joinpath("requirements")
    if not requirements_path.exists() and requirements_path.is_dir():
        requirements_path.mkdir(parents=False, exist_ok=False)

    return Config(
        cwd=cwd,
        requirements_dir=requirements_path,
        pyproject_file=path_to_pyproject_toml,
        caches=[cwd.joinpath(x) for x in cache_directories],
        artifacts=[cwd.joinpath(x) for x in artifacts_directories],
        venv_dir=cwd.joinpath(venv_dir),
        venv_activation_msg=(
            "\nTo activate the virtual env run:\n\n"
            f"On Windows, run:\n`{venv_dir}\\Scripts\\activate.bat`\n\n"
            f"On Unix or MacOS, run:\n`source {venv_dir}/bin/activate`\n"
        ),
        core_deps_alias=core_deps_alias,
        lock_file_format=lock_file_format,
        pyproject=tomli.loads(
            path_to_pyproject_toml.read_text(
                encoding="UTF-8",
            ),
        ),
    )



"""Library logger configuration."""
from __future__ import annotations

import sys

from loguru import logger


# https://github.com/Delgan/loguru/issues/109
def configure_logger() -> None:
    """Add custom configuration to the logger object."""
    logger.remove()
    fmt = (
        "<lvl>[{level}]</lvl> "
        "{message} "
        "<green>{name}:{function}:{line}</green> "
        "@ {time:HH:mm:ss}"
    )
    logger.add(sys.stderr, format=fmt)


configure_logger()

__all__ = [
    "logger",
]



"""test command utilitites."""
from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

import pytest

from pyrgo.core.utilities.text import colorize_text, path_to_lock_file

if TYPE_CHECKING:
    import sys

    if sys.version_info >= (3, 9):
        from typing import Literal
    else:
        from typing_extensions import Literal


@pytest.mark.unit()
@pytest.mark.parametrize(
    argnames=["color", "expected_text"],
    argvalues=[
        (
            "yellow",
            "\x1b[33mhi\x1b[0m",
        ),
        (
            "red",
            "\x1b[31mhi\x1b[0m",
        ),
    ],
)
def test_colorize_text(color: Literal["yellow", "red"], expected_text: str) -> None:
    """Test colorize text works as expected."""
    colorized_text = colorize_text(text="hi", color=color)
    assert colorized_text == expected_text


@pytest.mark.unit()
@pytest.mark.parametrize(
    argnames=[
        "base_path",
        "extra_path",
        "group",
        "lock_file_format",
        "expected_path",
    ],
    argvalues=[
        (
            pathlib.Path().cwd().joinpath("abc"),
            "xyz",
            "core",
            "txt",
            "xyz/core.txt",
        ),
    ],
)
def test_path_to_lock_file(
    base_path: pathlib.Path,
    extra_path: str,
    group: str,
    lock_file_format: str,
    expected_path: str,
) -> None:
    """Test path to lock file."""
    total_path = base_path.joinpath(extra_path)
    generated_path = path_to_lock_file(
        cwd=base_path,
        requirements_path=total_path,
        group=group,
        lock_file_format=lock_file_format,
    )
    assert generated_path == expected_path



"""Text utilities."""
from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    import sys

    if sys.version_info >= (3, 9):
        from typing import Literal
    else:
        from typing_extensions import Literal  # pragma: no cover
    from pathlib import Path


def colorize_text(
    *,
    text: str,
    color: Literal[
        "yellow",
        "red",
    ],
) -> str:
    """Colorize text."""
    return click.style(text=text, fg=color)


def path_to_lock_file(
    cwd: Path,
    requirements_path: Path,
    group: str,
    lock_file_format: str,
) -> str:
    """Parse to path to lock file."""
    path_diff = requirements_path.relative_to(cwd)
    return f"{path_diff!s}/{group}.{lock_file_format}"



"""sync command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.core import ops
from pyrgo.core.constants import app_config


@click.command()
@click.option(
    "-e",
    "--env",
    "environment",
    type=click.Choice(choices=app_config.available_envs),
    required=True,
    help="Sync to one of available enviroments.",
)
@click.option(
    "--editable/--no-editable",
    "editable",
    type=bool,
    default=True,
    show_default=True,
    help="Wheter or not to install current project in editable mode.",
)
def sync(*, environment: str, editable: bool) -> None:
    """Synchronize virtual environment with requirements.txt."""
    executed = ops.sync.execute(
        environment=environment,
        editable=editable,
        app_config=app_config,
    )
    if not isinstance(executed, Ok):
        sys.exit(1)
    sys.exit(0)



"""Root core CLI."""
from __future__ import annotations

import click

from pyrgo.cli.constants import PACKAGE_NAME
from pyrgo.cli.pyrgo_core.commands.add import add
from pyrgo.cli.pyrgo_core.commands.audit import audit
from pyrgo.cli.pyrgo_core.commands.build import build
from pyrgo.cli.pyrgo_core.commands.check import check
from pyrgo.cli.pyrgo_core.commands.clean import clean
from pyrgo.cli.pyrgo_core.commands.docs import docs
from pyrgo.cli.pyrgo_core.commands.fmt import fmt
from pyrgo.cli.pyrgo_core.commands.lock import lock
from pyrgo.cli.pyrgo_core.commands.remove import remove
from pyrgo.cli.pyrgo_core.commands.sync import sync
from pyrgo.cli.pyrgo_core.commands.test import test
from pyrgo.cli.pyrgo_core.commands.venv import venv
from pyrgo.cli.utils import add_commands


@click.group(
    context_settings={
        "help_option_names": [
            "-h",
            "--help",
        ],
    },
)
@click.version_option(
    None,
    "-v",
    "--version",
    package_name=PACKAGE_NAME,
    prog_name=PACKAGE_NAME,
)
def root() -> None:
    """Pyrgo. Python package manager."""


add_commands(
    cli=root,
    commands=[
        check,
        clean,
        fmt,
        lock,
        venv,
        add,
        test,
        docs,
        remove,
        build,
        sync,
        audit,
    ],
)



"""docs build documentation."""
from __future__ import annotations

import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.version_info >= (3, 9):
        from typing import Literal
    else:
        from typing_extensions import Literal

import click
from result import Ok

from pyrgo.core import ops


@click.command()
@click.option(
    "-t",
    "--theme",
    "theme",
    type=click.Choice(choices=["material", "mkdocs", "readthedocs"]),
    default="material",
    help="The theme to use when building your documentation.",
)
@click.option(
    "-s",
    "--strict",
    "strict",
    is_flag=True,
    default=False,
    help="Enable strict mode. This will cause MkDocs to abort the build on any warnings.",  # noqa: E501
)
def build(
    *,
    theme: Literal[
        "material",
        "mkdocs",
        "readthedocs",
    ],
    strict: bool,
) -> None:
    """Build project documentation."""
    executed = ops.docs.build.execute(
        theme=theme,
        strict=strict,
    )
    if not isinstance(executed, Ok):
        sys.exit(1)

    sys.exit(0)



"""docs new command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.core import ops
from pyrgo.core.constants import app_config


@click.command()
def new() -> None:
    """Create a new MkDocs project."""
    executed = ops.docs.new.execute(
        app_config=app_config,
    )
    if not isinstance(executed, Ok):
        sys.exit(1)

    sys.exit(0)



"""Cli utilitites."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import click


def add_commands(
    cli: click.Group,
    commands: list[click.Command],
) -> None:
    """Add commands to CLI group."""
    for command in commands:
        cli.add_command(cmd=command)



"""docs build operation."""
from __future__ import annotations

from typing import TYPE_CHECKING

from pyrgo.core.models.command import (
    PythonExecCommand,
)
from pyrgo.core.utilities.command import inform_and_run_program

if TYPE_CHECKING:
    import subprocess

    from result import Result


def execute(
    *,
    theme: str,
    strict: bool,
) -> Result[None, list[subprocess.CalledProcessError]]:
    """Execute docs build operation."""
    build_command = PythonExecCommand(
        program="mkdocs",
    ).add_args(
        args=[
            "build",
        ],
    )
    if strict:
        build_command.add_args(args=["--strict"])

    build_command.add_args(args=["--theme", theme])

    return inform_and_run_program(
        commands=[build_command],
    )



"""check operation."""
from __future__ import annotations

from typing import TYPE_CHECKING

from pyrgo.core.models.command import (
    PythonExecCommand,
)
from pyrgo.core.utilities.command import inform_and_run_program

if TYPE_CHECKING:
    import subprocess

    from result import Result

    from pyrgo.core.models.config import Config


def execute(
    *,
    add_noqa: bool,
    ignore_noqa: bool,
    app_config: Config,
) -> Result[None, list[subprocess.CalledProcessError]]:
    """Execute check operation."""
    ruff_command = PythonExecCommand(
        program="ruff",
    )
    mypy_command = PythonExecCommand(
        program="mypy",
    )

    if add_noqa:
        ruff_command.add_args(args=["--add-noqa"])
    if ignore_noqa:
        ruff_command.add_args(args=["--ignore-noqa"])

    for command in [ruff_command, mypy_command]:
        command.add_args(
            args=app_config.relevant_paths,
        )

    return inform_and_run_program(
        commands=[
            ruff_command,
            mypy_command,
        ],
    )



"""sync operation."""
from __future__ import annotations

from typing import TYPE_CHECKING

from pyrgo.core.models.command import (
    PythonExecCommand,
)
from pyrgo.core.utilities.command import inform_and_run_program
from pyrgo.core.utilities.text import path_to_lock_file

if TYPE_CHECKING:
    import subprocess

    from result import Result

    from pyrgo.core.models.config import Config


def execute(
    *,
    environment: str,
    editable: bool,
    app_config: Config,
) -> Result[None, list[subprocess.CalledProcessError]]:
    """Execute sync operation."""
    piptools_command = PythonExecCommand(
        program="piptools",
    )
    pip_command = PythonExecCommand(
        program="pip",
    )

    lock_file_path = path_to_lock_file(
        cwd=app_config.cwd,
        requirements_path=app_config.requirements_dir,
        group=environment,
        lock_file_format=app_config.lock_file_format,
    )

    piptools_command.add_args(
        args=[
            "sync",
            lock_file_path,
        ],
    )
    pip_command.add_args(
        args=[
            "install",
            "--no-deps",
        ],
    )
    if editable:
        pip_command.add_args(args=["-e"])

    pip_command.add_args(args=["."])

    return inform_and_run_program(
        commands=[
            piptools_command,
            pip_command,
        ],
    )



"""io utilities."""
from __future__ import annotations

from typing import TYPE_CHECKING
from zipfile import ZipFile

from result import Ok, Result

from pyrgo.core.resources import RESOURCER_PATH

if TYPE_CHECKING:
    from pathlib import Path


def prepare_starter_project(
    project_name: str,
    tmp_dir: Path,
) -> Result[Path, Exception]:
    """Unzip and prepare `starter-project` ready to me moved."""
    starter_project = "starter-project"
    with ZipFile(
        file=RESOURCER_PATH.joinpath(
            f"{starter_project}.zip",
        ),
    ) as starter_zip:
        starter_zip.extractall(tmp_dir)
        template_path = tmp_dir.joinpath(
            starter_project,
        )
        formatted_project_name = project_name.lower().replace("-", "_")
        template_path.joinpath("project_name").rename(
            target=template_path.joinpath(formatted_project_name),
        )
        template_path.joinpath("pyproject.toml").write_text(
            template_path.joinpath("pyproject.toml")
            .read_text(encoding="UTF-8")
            .replace("project_name", formatted_project_name),
        )

    return Ok(template_path)



"""command utilities."""
from __future__ import annotations

from typing import TYPE_CHECKING

import click
from result import Err, Ok, Result

if TYPE_CHECKING:
    import subprocess

    from pyrgo.core.models.command import PythonExecCommand
from pyrgo.core.utilities.text import colorize_text


def inform_and_run_program(
    commands: list[PythonExecCommand],
) -> Result[None, list[subprocess.CalledProcessError]]:
    """Inform and run program."""
    execution_failed = False
    subprocess_errors: list[subprocess.CalledProcessError] = []
    for command in commands:
        command_to_be_executed = " ".join(command.args)
        click.echo(
            message=colorize_text(
                text=f"Running {command_to_be_executed}",
                color="yellow",
            ),
        )
        executed = command.run()
        if not isinstance(executed, Ok):
            execution_failed = True
            subprocess_errors.append(executed.err())

    if execution_failed:
        return Err(subprocess_errors)

    return Ok()



"""venv operation."""
from __future__ import annotations

from typing import TYPE_CHECKING

import click
from result import Ok, Result

from pyrgo.core.models.command import (
    PythonExecCommand,
)
from pyrgo.core.utilities.command import inform_and_run_program

if TYPE_CHECKING:
    import subprocess

    from pyrgo.core.models.config import Config


def execute(app_config: Config) -> Result[None, list[subprocess.CalledProcessError]]:
    """Execute venv operation."""
    venv_command = PythonExecCommand(
        program="venv",
    ).add_args(
        args=[
            app_config.venv_dir.name,
        ],
    )
    if app_config.venv_dir.exists():
        if app_config.venv_dir.is_dir():
            click.echo(
                message="Project already has a virtual enviroment located"
                f" at `{app_config.venv_dir.name}`",
            )
            return Ok()
        if app_config.venv_dir.is_file():
            app_config.venv_dir.unlink()
            return inform_and_run_program(
                commands=[venv_command],
            )
        raise RuntimeError

    if not app_config.venv_dir.exists():
        return inform_and_run_program(
            commands=[venv_command],
        )
    raise RuntimeError
